/*************************************************
* Diffie-Hellman Source File                     *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/dh.h>
#include <opencl/numthry.h>

namespace OpenCL {

/*************************************************
* DH_PublicKey Constructor                       *
*************************************************/
DH_PublicKey::DH_PublicKey(const BigInt& prime, const BigInt& generator,
                           const BigInt& key) :
   p(prime), g(generator), y(key)
   {
   if(p <= 0 || g <= 1)
      throw Invalid_Argument("DH_PublicKey: Invalid prime or generator");
   }

/*************************************************
* Check Public DH Parameters                     *
*************************************************/
bool DH_PublicKey::check_params() const
   {
   if(g <= 1 || g >= p || y <= 1 || y >= p)
      return false;
   if(!verify_prime(p) || !verify_prime((p - 1) / 2))
      return false;
   return true;
   }

/*************************************************
* DH_PrivateKey Constructor                      *
*************************************************/
DH_PrivateKey::DH_PrivateKey(const BigInt& prime, const BigInt& generator,
                             const BigInt& x1, const BigInt& y1) :
   DH_PublicKey(prime, generator, y1), x(x1)
   {
   if(!x && y != 0)
      throw Invalid_Argument("DH_PrivateKey: Public key cannot be set "
                             "without private key");

   if(!x)
      x = random_integer(2 * dl_work_factor(p.bits()));
   powermod_x_p = FixedExponent_Exp(x, p);

   if(!y)
      y = powermod_x_p(g);
   }

/*************************************************
* Create a Diffie-Hellman Key                    *
*************************************************/
DH_PrivateKey::DH_PrivateKey(u32bit bits)
   {
   if(bits < 64)
      throw Invalid_Argument("DH: Can't make a key that is only " +
                             to_string(bits) + " bits long");

   g = 2;
   p = random_safe_prime(bits);
   x = random_integer(2 * dl_work_factor(p.bits()));
   powermod_x_p = FixedExponent_Exp(x, p);
   y = powermod_x_p(x);
   }

/*************************************************
* Check Private DH Parameters                    *
*************************************************/
bool DH_PrivateKey::check_params() const
   {
   if(!DH_PublicKey::check_params())
      return false;
   if(y != powermod_x_p(g))
      return false;
   return true;
   }

/*************************************************
* Derive a key                                   *
*************************************************/
SecureVector<byte> DH_PrivateKey::derive_key(const byte w[],
                                             u32bit w_len) const
   {
   return derive_key(decode(w, w_len));
   }

/*************************************************
* Derive a key                                   *
*************************************************/
SecureVector<byte> DH_PrivateKey::derive_key(const DH_PublicKey& key) const
   {
   return derive_key(key.get_y());
   }

/*************************************************
* Derive a key                                   *
*************************************************/
SecureVector<byte> DH_PrivateKey::derive_key(const BigInt& w) const
   {
   return encode(powermod_x_p(w));
   }

}
