/*************************************************
* DSA Source File                                *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/dsa.h>
#include <opencl/numthry.h>
#include <opencl/primes.h>

namespace OpenCL {

/*************************************************
* DSA_PublicKey Constructor                      *
*************************************************/
DSA_PublicKey::DSA_PublicKey(const BigInt& prime, const BigInt& q1,
                             const BigInt& generator, const BigInt& y1) :
   p(prime), q(q1), g(generator), y(y1), powermod_g_p(g, p), powermod_y_p(y, p)
   {
   if(p <= 0 || q <= 0 || g <= 1)
      throw Invalid_Argument("DSA_PublicKey: Invalid prime or generator");
   }

/*************************************************
* Check Public DSA Parameters                    *
*************************************************/
bool DSA_PublicKey::check_params() const
   {
   if((p - 1) % q != 0)
      return false;
   if(g <= 1 || g >= p)
      return false;
   if(!verify_prime(p) || !verify_prime(q))
      return false;
   return true;
   }

/*************************************************
* DSA Verification Function                      *
*************************************************/
bool DSA_PublicKey::verify(const byte msg[], u32bit msg_len,
                           const byte sig[], u32bit sig_len) const
   {
   if(sig_len != 2*q.bytes())
      return false;

   t1.binary_decode(msg, msg_len);
   r.binary_decode(sig, sig_len / 2);
   s.binary_decode(sig + sig_len / 2, sig_len / 2);

   if(r.is_zero() || s.is_zero() || r >= q || s >= q)
      return false;

   BigInt w = inverse_mod(s, q);
   BigInt u1 = (w * t1) % q;
   BigInt u2 = (w * r) % q;
   BigInt v = (powermod_g_p(u1) * powermod_y_p(u2)) % p;
   return (v % q == r);
   }

/*************************************************
* DSA_PrivateKey Constructor                     *
*************************************************/
DSA_PrivateKey::DSA_PrivateKey(const BigInt& prime, const BigInt& q1,
                               const BigInt& generator, const BigInt& x1,
                               const BigInt& y1) :
   DSA_PublicKey(prime, q1, generator, y1), x(x1)
   {
   if(!x && y != 0)
      throw Invalid_Argument("DSA_PrivateKey: Public key cannot be set "
                             "without private key");

   if(!x)
      x = random_integer(q.bits() - 1);
   if(!y)
      {
      y = powermod_g_p(x);
      powermod_y_p = FixedBase_Exp(y, p);
      }
   }

/*************************************************
* Create a DSA Key                               *
*************************************************/
DSA_PrivateKey::DSA_PrivateKey(u32bit pbits, bool use_fips_keygen)
   {
   if(pbits < 512 || pbits % 8)
      throw Invalid_Argument("DSA_PrivateKey: Invalid size " +
                             to_string(pbits) + " for prime p");

   if(use_fips_keygen && (!valid_fips186_p_size(pbits)))
      throw Invalid_Argument("DSA_PrivateKey: Size of p not legal for "
                             "FIPS-186 prime generation");

   if(use_fips_keygen)
      generate_dsa_primes(p, q, pbits);
   else
      {
      u32bit qbits = 2 * dl_work_factor(pbits);
      q = random_prime(qbits);
      BigInt q_2 = 2 * q;
      do
         p = random_integer(pbits - qbits - 1) * q_2 + 1;
      while(p.bits() != pbits || !is_prime(p));
      }

   keygen_operation();
   }

/*************************************************
* Create a DSA Key from a seed                   *
*************************************************/
DSA_PrivateKey::DSA_PrivateKey(const byte seed[], u32bit seed_length,
                               u32bit pbits, u32bit start)
   {
   if(!generate_dsa_primes(p, q, seed, seed_length, pbits, start))
      throw Invalid_Argument("DSA_PrivateKey: The seed value given does not "
                             "generate a DSA key");
   keygen_operation();
   }

/*************************************************
* Handle Parts of DSA Key Generation             *
*************************************************/
void DSA_PrivateKey::keygen_operation()
   {
   BigInt e = (p - 1) / q;
   u32bit prime_no = 0;
   do
      g = power_mod(PRIMES[prime_no++], e, p);
   while(g == 1);
   powermod_g_p = FixedBase_Exp(g, p);

   x = random_integer(q.bits() - 1);
   y = powermod_g_p(x);
   powermod_y_p = FixedBase_Exp(y, p);
   }

/*************************************************
* Check Private DSA Parameters                   *
*************************************************/
bool DSA_PrivateKey::check_params() const
   {
   if(!DSA_PublicKey::check_params())
      return false;
   if(y != powermod_g_p(x))
      return false;
   return true;
   }

/*************************************************
* Check for a p size authorized by FIPS 186-2    *
*************************************************/
bool DSA_PrivateKey::valid_fips186_p_size(u32bit pbits) const
   {
   return ((pbits % 64 == 0) || (pbits >= 512) || (pbits <= 1024));
   }

/*************************************************
* DSA Signature Operation                        *
*************************************************/
SecureVector<byte> DSA_PrivateKey::sign(const byte in[], u32bit len) const
   {
   t1.binary_decode(in, len);

   while(true)
      {
      BigInt k(BigInt::Random, q.bits());
      while(k >= q)
         k.randomize(q.bits());

      r = powermod_g_p(k) % q;
      s = (inverse_mod(k, q) * mul_add(x, r, t1)) % q;

      if(r.is_zero() || s.is_zero())
         continue;

      SecureVector<byte> output(2*q.bytes());
      r.binary_encode(output + (q.bytes() - r.bytes()));
      s.binary_encode(output + output.size() / 2 + (q.bytes() - s.bytes()));
      return output;
      }
   }

}
